# Works like Configuration:
# Inherit from the DataUser class to access data of all kinds (listed above)

import rdflib
from rdflib import Graph, Namespace, ConjunctiveGraph
from rdflib.namespace import NamespaceManager
from .quantity import Quantity
from datetime import datetime as DT
import datetime
import os
import logging
from .configure import Configureable, Configuration, ConfigValue

__all__ = [
    "Data",
    "RDFSource",
    "SerializationSource",
    "TrixSource",
    "SPARQLSource",
    "DefaultSource"]

L = logging.getLogger(__name__)


class Data(Configuration, Configureable):

    """
    Provides configuration for access to the database.

    Usally doesn't need to be accessed directly
    """

    def __init__(self, conf=False):
        Configuration.__init__(self)
        Configureable.__init__(self)
        if conf is not False:
            self.init_conf = conf
        else:
            self.init_conf = Configureable.conf

        # We copy over all of the configuration that we were given
        self.copy(self.init_conf)
        ns_string = self.get(
            'rdf.namespace',
            'http://example.org/TestNamespace/entities/')

        self['rdf.namespace'] = Namespace(ns_string)
        self.namespace = self['rdf.namespace']

        # TODO: Add support for defining additional data types from the configs
        self.dt_ns = Namespace(self.namespace["datatypes"] + "/")
        quant_datatype = self.dt_ns["quantity"]
        if quant_datatype not in rdflib.term._toPythonMapping:
            rdflib.term.bind(quant_datatype, Quantity, Quantity.parse)
        self.sources = dict()
        self.register_source(DefaultSource)

    @classmethod
    def open(cls, file_name):
        """ Open a file storing configuration in a JSON format """
        Configureable.setConf(Configuration.open(file_name))
        return cls()

    def openDatabase(self):
        """ Open a the configured database """
        self._init_rdf_graph()
        L.debug("opening " + str(self.source))
        self.source.open()
        nm = NamespaceManager(self['rdf.graph'])
        self['rdf.namespace_manager'] = nm
        self['rdf.graph'].namespace_manager = nm

        nm.bind("dt", self.dt_ns)
        nm.bind("", self['rdf.namespace'])

        # TODO: Extract classes recorded in the graph
        #       First, look at the :pythonClass attribute attached to the RDF class resource.
        #       - Look up the class in the PythonClassRegistry using the ID
        #         attached at the :pythonClass attribute.
        #       - Verify the python class name in the registry
        #       - Get the python module location (whether remote or local)
        #       - Load the python module (downloading it if necessary) and add it to the yarom
        #         namespace
        # TODO: Add support for loading python packages with setuptools.

    def register_source(self, source):
        if source.name in self.sources:
            raise Exception(
                "There is already an rdf source with the name " +
                str(source.name))
        else:
            self.sources[source.name] = source

    def closeDatabase(self):
        """ Close a the configured database """
        self.source.close()

    def _init_rdf_graph(self):
        # Set these in case they were left out
        c = self.init_conf
        self['rdf.source'] = c['rdf.source'] = c.get('rdf.source', 'default')
        self['rdf.store'] = c['rdf.store'] = c.get('rdf.store', 'default')
        self['rdf.store_conf'] = c['rdf.store_conf'] = c.get(
            'rdf.store_conf',
            'default')

        source_graph = self.sources[self['rdf.source'].lower()]()
        self.source = source_graph

        if self.get("rdf.inference", False):
            if 'rdf.rules' not in self:
                self['rdf.inference'] = False
                raise Exception(
                    "You've set `rdf.inference' in your configuration. Please provide n3 rules in your configuration (property name `rdf.rules') as well in order to use rdf inference.")

            import warnings
            warnings.filterwarnings(
                'ignore',
                "Missing pydot library")  # Filters an obnoxious warning from FuXi
            # Filters a warning from rdflib not closing its files from a parse
            warnings.filterwarnings(
                'ignore',
                ".*unclosed file <_io.BufferedReader .*")
            try:
                from FuXi.Rete.RuleStore import SetupRuleStore
                from FuXi.Rete.Util import generateTokenSet
                from FuXi.Horn.HornRules import HornFromN3
            except ImportError:
                self['rdf.inference'] = False
                raise Exception(
                    "You've set `rdf.inference' in your configuration, but you do not have FuXi installed, so inference cannot be performed.")

            # fetch the derived object's graph
            rule_store, rule_graph, network = SetupRuleStore(makeNetwork=True)

            # build a network of rules
            for rule in HornFromN3(self['rdf.rules']):
                network.buildNetworkFromClause(rule)

            def infer(graph, new_data):
                """ Fire FuXi rule engine to infer triples """
                # apply rules to original facts to infer new facts
                closureDeltaGraph = Graph()
                network.inferredFacts = closureDeltaGraph

                network.feedFactsToAdd(generateTokenSet(new_data))
                # combine original facts with inferred facts
                if graph:
                    for x in closureDeltaGraph:
                        graph.add(x)
            self['fuxi.network'] = network
            self['fuxi.rule_graph'] = rule_graph
            self['fuxi.rule_store'] = rule_store
            self['fuxi.infer_func'] = infer

            # XXX: Not sure if this is the most appropriate way to set
            #      up the network
            source_graph._get = source_graph.get

            def get():
                """ A one-time wrapper. Resets to the actual `get` after being called once """
                g = source_graph._get()  # get the graph in the normal way
                infer(False, g)  # add the initial facts to the rete network
                source_graph.get = source_graph._get  # restore the old `get`
                return g

            source_graph.get = get

        self.link('semantic_net_new', 'semantic_net', 'rdf.graph')
        self['rdf.graph'] = source_graph
        return source_graph


def modification_date(filename):
    t = os.path.getmtime(filename)
    return datetime.datetime.fromtimestamp(t)


class RDFSource(ConfigValue, Configureable):

    """ Base class for data sources.

    Alternative sources should dervie from this class
    """

    def __init__(self, **kwargs):
        super(RDFSource, self).__init__(**kwargs)
        self.graph = False

    def get(self):
        if self.graph is False:
            raise Exception(
                "Must call openDatabase on Data object before using the database")
        return self.graph

    def close(self):
        if self.graph is False:
            return
        self.graph.close()
        self.graph = False

    def open(self):
        """ Called on ``yarom.connect()`` to set up and return the rdflib graph.
        Must be overridden by sub-classes.
        """
        raise NotImplementedError()

    def __repr__(self):
        return self.__class__.__name__ + "()"

class SerializationSource(RDFSource):

    """ Reads from an RDF serialization or, if the configured database is more recent, then from that.

        The database store is configured with::

            "rdf.source" = "serialization"
            "rdf.store" = <your rdflib store name here>
            "rdf.serialization" = <your RDF serialization>
            "rdf.serialization_format" = <your rdflib serialization format used>
            "rdf.store_conf" = <your rdflib store configuration here>

    """
    name = 'serialization'

    def open(self):
        if not self.graph:
            self.graph = True
            import glob
            # Check the ages of the files. Read the more recent one.
            g0 = ConjunctiveGraph(store=self.conf['rdf.store'])
            database_store = self.conf['rdf.store_conf']
            source_file = self.conf['rdf.serialization']
            file_format = self.conf['rdf.serialization_format']
            # store_time only works for stores that are on the local
            # machine.
            try:
                store_time = modification_date(database_store)
                # If the store is newer than the serialization
                # get the newest file in the store
                for x in glob.glob(database_store + "/*"):
                    mod = modification_date(x)
                    if store_time < mod:
                        store_time = mod
            except:
                store_time = DT.min

            trix_time = modification_date(source_file)

            g0.open(database_store, create=True)

            if store_time > trix_time:
                # just use the store
                pass
            else:
                # delete the database and read in the new one
                # read in the serialized format
                import warnings
                # Filters a warning from rdflib not closing its files from a
                # parse
                warnings.filterwarnings(
                    'ignore',
                    ".*unclosed file <_io.BufferedReader .*")
                g0.parse(source_file, format=file_format)

            self.graph = g0

        return self.graph


class TrixSource(SerializationSource):

    """ A SerializationSource specialized for TriX

        The database store is configured with::

            "rdf.source" = "trix"
            "rdf.trix_location" = <location of the TriX file>
            "rdf.store" = <your rdflib store name here>
            "rdf.store_conf" = <your rdflib store configuration here>

    """

    name = 'trix'

    def __init__(self, **kwargs):
        super(TrixSource, self).__init__(**kwargs)
        h = self.conf.get('trix_location', 'UNSET')
        self.conf.link('rdf.serialization', 'trix_location')
        self.conf['rdf.serialization'] = h
        self.conf['rdf.serialization_format'] = 'trix'


class SPARQLSource(RDFSource):

    """ Reads from and queries against a remote data store

        Configure like::

            "rdf.source" = "sparql_endpoint"
    """
    name = 'sparql_endpoint'

    def open(self):
        # XXX: If we have a source that's read only, should we need to set the
        # store separately??
        g0 = ConjunctiveGraph('SPARQLUpdateStore')
        g0.open(tuple(self.conf['rdf.store_conf']))
        self.graph = g0
        return self.graph


class DefaultSource(RDFSource):

    """ Reads from and queries against a configured database.

        The default configuration.

        The database store is configured with::

            "rdf.source" = "default"
            "rdf.store" = <your rdflib store name here>
            "rdf.store_conf" = <your rdflib store configuration here>

        Leaving unconfigured simply gives an in-memory data store.
    """
    name = 'default'

    def open(self):
        self.graph = ConjunctiveGraph(self.conf['rdf.store'])
        self.graph.open(self.conf['rdf.store_conf'], create=True)


class _UTC(datetime.tzinfo):

    """UTC"""

    ZERO = datetime.timedelta(0)

    def utcoffset(self, dt):
        return _UTC.ZERO

    def tzname(self, dt):
        return "UTC"

    def dst(self, dt):
        return _UTC.ZERO
utc = _UTC()


class _B(ConfigValue):

    def __init__(self, f):
        self.v = False
        self.f = f

    def get(self):
        if not self.v:
            self.v = self.f()

        return self.v

    def invalidate(self):
        self.v = False



from __future__ import print_function

import logging
import six
import rdflib as R
from .dataUser import DataUser
from .mappedProperty import MappedPropertyClass
from .utils import slice_dict

L = logging.getLogger(__name__)

ProplistToProptype = {"datatypeProperties": "DatatypeProperty",
                      "objectProperties": "ObjectProperty",
                      "_": "UnionProperty"}


class MappedClass(type):

    """A type for MappedClasses

    Sets up the graph with things needed for MappedClasses
    """
    def __init__(self, name, bases, dct):
        L.debug("INITIALIZING %s", name)
        super(MappedClass, self).__init__(name, bases, dct)
        if 'auto_mapped' in dct:
            self.mapped = True
        else:
            self.mapped = False

        self.__rdf_type = None
        # Set the rdf_type early
        if 'rdf_type' in dct:
            self.__rdf_type = dct['rdf_type']

        if self.__rdf_type is None:
            self.__rdf_type = self.base_namespace[self.__name__]

        self.__rdf_namespace = None
        rdf_ns = dct.get('rdf_namespace', None)
        if rdf_ns is not None:
            L.debug("Setting rdf_namespace to {}".format(rdf_ns))
            self.__rdf_namespace = rdf_ns

        if self.__rdf_namespace is None:
            L.debug("Setting rdf_namespace to {}".format(self.base_namespace[self.__name__]))
            self.__rdf_namespace = R.Namespace(
                self.base_namespace[self.__name__] + "/")

        self._du = None
        self.dataObjectProperties = []
        self.children = []
        for x in bases:
            try:
                self.dataObjectProperties += x.dataObjectProperties
            except AttributeError:
                pass

    @property
    def rdf_type(self):
        return self.__rdf_type

    @rdf_type.setter
    def rdf_type(self, new_type):
        if not isinstance(new_type, R.URIRef) and \
                isinstance(new_type, (str, six.text_type)):
            new_type = R.URIRef(new_type)
        self.__rdf_type = new_type

    @property
    def rdf_namespace(self):
        return self.__rdf_namespace

    @classmethod
    def make_class(
            cls,
            name,
            bases,
            objectProperties=False,
            datatypeProperties=False):
        """
        Intended to be used for setting up a class from the RDF graph, for
        instance.
        """
        # Need to distinguish datatype and object properties...
        if not datatypeProperties:
            datatypeProperties = []
        if not objectProperties:
            objectProperties = []
        cls(name, bases,
            dict(objectProperties=objectProperties,
                 datatypeProperties=datatypeProperties))

    @property
    def du(self):
        # This is just a little indirection to make sure we initialize the
        # DataUser property before using it. Initialization isn't done in
        # __init__ because I wanted to make sure you could call `connect`
        # before or after declaring your classes
        if not self._du:
            self._du = DataUser()
        return self._du

    @du.setter
    def du(self, value):
        assert(isinstance(value, DataUser))
        self._du = value

    def __lt__(self, other):
        res = False
        if issubclass(other, self) and not issubclass(self, other):
            res = True
        elif issubclass(self, other) == issubclass(other, self):
            res = self.__name__ < other.__name__
        return res

    def on_mapper_add_class(self, mapper):
        """ Called by :class:`yarom.mapper.Mapper`

        Registers certain properties of the class
        """
        self.mapper = mapper
        L.debug("REGISTERING %s", self.__name__)
        parents = self.__bases__
        mapped_parents = tuple(x for x in parents
                               if isinstance(x, MappedClass))

        for c in mapped_parents:
            c.add_child(self)

        self.parents = mapped_parents

        self.addProperties('objectProperties')
        self.addProperties('datatypeProperties')
        self.addProperties('_')

        return self

    def after_mapper_module_load(self, mapper):
        """ Called after all classes in a module have been loaded """

    def on_mapper_remove_class(self, mapper):
        L.debug("DEREGISTERING %s", self.__name__)

        for c in self.parents:
            c.remove_child(self)

    def remove_child(self, child):
        if hasattr(self, 'children'):
            L.debug('removing child %s@0x%x %s@0x%x',
                    self, id(self), child, id(child))
            self.children.remove(child)
        else:
            raise Exception(
                "Cannot remove child {0} from {1} as {1} has not yet "
                "been registered".format(
                    child,
                    self))

    def add_child(self, child):
        if hasattr(self, 'children'):
            L.debug('adding child %s@0x%x %s@0x%x',
                    self, id(self), child, id(child))
            self.children.append(child)
        else:
            raise Exception(
                "Cannot add child {0} to {1} "
                "as {1} has not yet been registered".format(
                    child,
                    self))

    def map(self):
        """
        Maps the class to the configured rdf graph.
        """
        # NOTE: Map should be quick: it runs for every DataObject sub-class
        #       created and possibly several times in testing
        L.debug("MAPPING %s", self.__name__)
        TypeDataObject = \
            self.mapper.lookup_class('yarom.dataObject.TypeDataObject')
        TypeDataObject.mapped = True

        self.rdf_type_object = TypeDataObject(ident=self.rdf_type)

        self._add_parents_to_graph()
        self.mapped = True

        return self

    def unmap(cls):
        """
        Unmaps the class
        """
        L.debug("UNMAPPING %s", cls.__name__)
        del cls.mapper.RDFTypeTable[cls.__rdf_type]
        # XXX: What else to do here?

    def _remove_namespace_from_manager(cls):
        # XXX: Only way I can think to do this is to remake the ns manager
        pass

    def _add_namespace_to_manager(self):
        self.du['rdf.namespace_manager'].bind(
            self.__name__,
            self.__rdf_namespace,
            replace=True)

    def _add_parents_to_graph(self):
        m = self.mapper.modules['yarom.dataObject']
        L.debug('adding parents for %s', self)
        for parent in self.parents:
            ancestors = (x for x in parent.mro()
                         if issubclass(x, m.DataObject))
            for ancestor in ancestors:
                self.rdf_type_object.relate(
                    'rdfs_subClassOf',
                    ancestor.rdf_type_object,
                    m.RDFSSubClassOfProperty)

    def addProperties(cls, listName):
        # TODO: Make an option string to abbreviate these options
        if hasattr(cls, listName):
            propList = getattr(cls, listName)
            propType = ProplistToProptype[listName]

            assert(isinstance(propList, (tuple, list, set)))
            for x in propList:
                value_type = None
                p = None
                if isinstance(x, tuple):
                    if len(x) > 2:
                        value_type = x[2]
                    name = x[0]
                    p = _create_property(
                        cls,
                        name,
                        propType,
                        value_type=value_type)
                elif isinstance(x, dict):
                    if 'prop' in x:
                        p = cls.mapper.MappedClasses[x['prop']]
                    else:
                        name = x['name']
                        del x['name']

                        if 'type' in x:
                            value_type = x['type']
                            del x['type']

                        p = _create_property(
                            cls,
                            name,
                            propType,
                            value_type=value_type,
                            **x)
                else:
                    p = _create_property(cls, x, propType)
                cls.dataObjectProperties.append(p)
            setattr(cls, '_' + listName, propList)
            setattr(cls, listName, [])

    def _cleanupGraph(cls):
        """
        Cleans up the graph by removing statements that can't be connected to
        typed statement.
        """
        # XXX: This might belong in DataUser instead
        q = """DELETE { ?b ?x ?y }
        WHERE
        {
            ?b ?x ?y .
            FILTER (NOT EXISTS { ?b rdf:type ?c } ) .
        }"""
        cls.du.rdf.update(q)


def _create_property(
        owner_type,
        linkName,
        property_type,
        value_type=None,
        multiple=True,
        link=None):
    # XXX: This should actually get called for all of the properties when their
    #      owner classes are defined. The initialization, however, must happen
    #      with the owner object's creation
    mapper = owner_type.mapper
    ysp = mapper.load_module('yarom.simpleProperty')
    do = mapper.load_module('yarom.dataObject')

    properties = slice_dict(locals(), ['owner_type', 'linkName', 'multiple'])

    owner_class_name = owner_type.__name__
    property_class_name = owner_class_name + "_" + linkName

    x = None
    if property_type == 'ObjectProperty':
        x = ysp.ObjectProperty
        if value_type is None:
            value_type = do.DataObject
        properties['value_type'] = value_type
    elif property_type == 'DatatypeProperty':
        x = ysp.DatatypeProperty
    else:
        x = ysp.UnionProperty

    if link is not None:
        properties['link'] = link
    elif (hasattr(owner_type, 'rdf_namespace') and
            owner_type.rdf_namespace is not None):
        properties['link'] = owner_type.rdf_namespace[linkName]
    c = MappedPropertyClass(property_class_name, (x,), properties)
    mapper.add_class(c)
    return c



# -*- coding: utf-8 -*-

"""
Most statements correspond to some action on the database. Some of these
actions may be complex, but intuitively ``a.B()``, the Query form, will query
against the database for the value or values that are related to ``a`` through
``B``; on the other hand, ``a.B(c)``, the Update form, will add a statement to
the database that ``a`` relates to ``c`` through ``B``. For the Update form, a
Relationship object describing the relationship stated is returned as a side-
effect of the update.

The Update form can also be accessed through the set() method of a Property and
the Query form through the get() method like::

    a.B.set(c)

and::

    a.B.get()

The get() method also allows for parameterizing the query in ways specific to
the Property.

Classes
-------
.. automodule:: yarom.dataObject
.. automodule:: yarom.dataUser
.. automodule:: yarom.data
.. automodule:: yarom.zodb
.. automodule:: yarom.configure
.. automodule:: yarom.graphObject
.. automodule:: yarom.mapper
"""


import logging

from .mapper import Mapper

__version__ = '0.11.0.post0'
__author__ = 'Mark Watts'

this_module = __import__('yarom')
this_module.connected = False

L = logging.getLogger(__name__)

DEFAULT_MODULES_TO_LOAD = ["yarom.dataObject",
                           "yarom.objectCollection",
                           "yarom.relationship",
                           "yarom.classRegistry"]

__all__ = ['setConf',
           'config',
           'loadConfig',
           'loadData',
           'connect',
           'disconnect']


MAPPER = None


def yarom_import(cname_or_mname, cnames=None):
    global MAPPER
    if cnames:
        mpart = cname_or_mname
    else:
        mpart, cpart = cname_or_mname.rsplit('.', 1)
        cnames = (cpart,)
    m = MAPPER.load_module(mpart)
    if len(cnames) == 1:
        return getattr(m, cnames[0])
    else:
        return tuple(getattr(m, cname) for cname in cnames)


def yarom_dependency(mname):
    global MAPPER
    MAPPER.add_module_dependency(mname)


def config(key=None, value=None):
    from .configure import Configureable
    if key is None:
        return Configureable.conf
    elif value is None:
        return Configureable.conf[key]
    else:
        Configureable.conf[key] = value


def loadConfig(f):
    """ Load configuration for the module """
    from .configure import Configureable
    from .data import Data
    Configureable.setConf(Data.open(f))
    return Configureable.conf


def disconnect(c=False):
    """ Close the database """
    from .configure import Configureable
    global MAPPER
    m = this_module
    if not m.connected:
        return

    if not c:
        c = Configureable.conf
    MAPPER.deregister_all()  # NOTE: We do NOT unmap on disconnect
    # Note that `c' could be set in one of the previous branches;
    # don't try to simplify this logic.
    if c:
        c.closeDatabase()
    MAPPER = None
    m.connected = False


def loadData(data, dataFormat):
    import rdflib
    if isinstance(data, str):
        config('rdf.graph').parse(data, format=dataFormat)
    elif isinstance(data, rdflib.ConjunctiveGraph):
        g = config('rdf.graph')
        for x in data.quads((None, None, None, None)):
            g.add(x)


def connect(conf=False,
            do_logging=False,
            data=False,
            dataFormat='n3',
            modulesToLoad=None):
    """Load desired configuration and open the database

    Parameters
    ----------
    conf : :class:`str`, :class:`Data <yarom.data.Data>`, :class:`Configuration <yarom.configure.Configuration>` or :class:`dict`, optional
        The configuration for the YAROM connection
    do_logging : bool, optional
        If True, turn on debug level logging. The default is False.
    data : str or rdflib.ConjunctiveGraph, optional
        If provided, specifies a file to load into the library.
    dataFormat : str, optional
        If provided, specifies the file format of the file pointed specified by
        `data`.

        The formats available are those accepted by RDFLib's serializer
        plugins. 'n3' is the default.
    """
    from .configure import Configureable
    from .data import (
        SPARQLSource,
        TrixSource,
        SerializationSource)
    import atexit
    global MAPPER
    if MAPPER is None:
        MAPPER = Mapper(('yarom.dataObject.DataObject',
                         'yarom.simpleProperty.SimpleProperty'))
    m = this_module
    if m.connected:
        print("yarom already connected")
        return

    if do_logging:
        logging.basicConfig(level=logging.DEBUG)
    if modulesToLoad is None:
        modulesToLoad = DEFAULT_MODULES_TO_LOAD

    setConf(conf)
    dbconn = Configureable.conf
    dbconn.register_source(SPARQLSource)
    dbconn.register_source(TrixSource)
    dbconn.register_source(SerializationSource)

    dbconn.openDatabase()
    L.info("Connected to database")

    atexit.register(disconnect)

    for mod in modulesToLoad:
        MAPPER.load_module(mod)

    MAPPER.remap()
    m.connected = True
    if data:
        loadData(data, dataFormat)


def setConf(conf):
    """ Set the configuration

    Parameters
    ----------
    conf : str, Data, Configuration or dict, optional
        The configuration to load.

        If a Data object is provided, then it's used as is for the
        configuration.

        If either a Python dict or a Configuration object are provided, then the
        contents of that object is used to make a Data object for configuration.

        If a string is provided, then the file is read in as JSON to be parsed as
        a dict and from there is treated as if you had passed that dict to
        connect.

        The default action is to attempt to open a file called 'yarom.conf' from
        your current directory as the configuration. Failing that, an 'empty'
        config with default values will be loaded.
    """
    from .configure import Configuration, Configureable
    from .data import Data
    if conf:
        if isinstance(conf, Data):
            Configureable.setConf(conf)
        elif isinstance(conf, (Configuration, dict)):
            Configureable.setConf(Data(conf))
        elif isinstance(conf, str):
            Configureable.setConf(Data.open(conf))
    else:
        try:
            Configureable.setConf(Data.open("yarom.conf"))
        except Exception:
            L.info("Couldn't load default configuration")
            Configureable.setConf(Data())



from rdflib import Graph, Namespace
from rdflib.namespace import RDF, NamespaceManager
import logging
from .configure import Configureable
from .data import Data
from .rdfUtils import triples_to_bgp

L = logging.getLogger(__name__)

__all__ = ["DataUser"]


class DataUser(Configureable):
    """ A convenience wrapper for users of the database

    Classes which use the database should inherit from DataUser.
    """

    # TODO: Make these metadata accessible for configuration managers
    # TODO: Illustrate how setting variables in a config file translates into
    #       these configurations
    configuration_variables = {
            "rdf.namespace" : {
                "description" : "The base namespace for user objects and information.",
                "type" : Namespace,
                "directly_configureable" : True
                },
            "rdf.store" : {
                "description" : "Influences how statements are added to the graph.",
                "type" : str,
                "directly_configureable" : True
                },
            "rdf.graph" : {
                "description" : "The rdflib graph object used for querying and storage of user objects and information.",
                "type" : Graph
                },
            "rdf.namespace_manager" : {
                "description" : "The namespace manager associated with the rdf.graph. Stores prefixes that get used by yarom",
                "type" : NamespaceManager
                }
            }

    def __init__(self, **kwargs):
        super(DataUser, self).__init__(**kwargs)

        if not isinstance(self.conf, Data):
            Configureable.setConf(Data())

    @property
    def rdf(self):
        return self.conf['rdf.graph']

    @rdf.setter
    def rdf(self, value):
        self.conf['rdf.graph'] = value

    @property
    def namespace_manager(self):
        return self.conf['rdf.namespace_manager']

    def _remove_from_store(self, g):
        # Note the assymetry with _add_to_store. You must add actual elements,
        # but deletes can be performed as a query
        for group in grouper(g, 1000):
            temp_graph = Graph()
            for x in group:
                if x is not None:
                    temp_graph.add(x)
                else:
                    break
            s = " DELETE DATA {" + temp_graph.serialize(format="nt") + " } "
            L.debug("deleting. s = " + s)
            self.conf['rdf.graph'].update(s)

    def _add_to_store(self, g, graph_name=False):
        if self.conf['rdf.store'] == 'SPARQLUpdateStore':
            # XXX With Sesame, for instance, it is probably faster to do a PUT
            #     with over the endpoint's rest interface. Just need to do it
            #     for some common endpoints

            try:
                gs = g.serialize(format="nt")
            except:
                gs = triples_to_bgp(g)

            if graph_name:
                s = " INSERT DATA { GRAPH "+graph_name.n3()+" {" + gs + " } } "
            else:
                s = " INSERT DATA { " + gs + " } "
                L.debug("update query = " + s)
                self.conf['rdf.graph'].update(s)
        else:
            gr = self.conf['rdf.graph']
            for x in g:
                gr.add(x)
            if self.conf.get('rdf.inference', False):
                self.conf['fuxi.infer_func'](gr, g)

        if self.conf['rdf.source'] == 'ZODB':
            import transaction
            # Commit the current commit
            transaction.commit()
            # Fire off a new one
            transaction.begin()

    def retract_statements(self, statements):
        """
        Remove a set of statements from the database.

        Parameters
        ----------
        triples : iter of (:class:`rdflib.term.URIRef`, :class:`rdflib.term.URIRef`, :class:`rdflib.term.URIRef`)
            A set of triples to remove
        """
        for x in statements:
            self.rdf.remove(x)

    def _remove_from_store_by_query(self, q):
        import logging as L
        s = " DELETE WHERE {" + q + " } "
        L.debug("deleting. s = " + s)
        self.conf['rdf.graph'].update(s)

    def add_statements(self, graph):
        """
        Add a set of statements to the database.

        Annotates the addition with uploader name, etc

        Parameters
        ----------
        triples : iter of (:class:`rdflib.term.URIRef`, :class:`rdflib.term.URIRef`, :class:`rdflib.term.URIRef`)
            A set of triples to add to the graph
        """
        self._add_to_store(graph)

    def _reify(self, g, s):
        """
        Add a statement object to g that binds to s
        """
        n = self.conf['new_graph_uri'](s)
        g.add((n, RDF['type'], RDF['Statement']))
        g.add((n, RDF['subject'], s[0]))
        g.add((n, RDF['predicate'], s[1]))
        g.add((n, RDF['object'], s[2]))
        return n


def grouper(iterable, n):
    """Collect data into fixed-length chunks or blocks"""
    # grouper('ABCDEFG', 3) --> ABC DEF G
    args = [iter(iterable)] * n
    while True:
        l = []
        try:
            for x in args:
                l.append(next(x))
        except StopIteration:
            pass
        yield l
        if len(l) < n:
            break



from __future__ import print_function

# Directions for traversal across triples
UP = 'up'
DOWN = 'down'


def print_graph(g, hide_namespaces=False):
    s = g.serialize(format='n3').decode("UTF-8")
    if hide_namespaces:
        lines = s.splitlines()
        s = "\n".join(l for l in lines if not l.startswith("@prefix"))
    print(s)


def serialize_rdflib_term(x, namespace_manager=None):
    return x.n3(namespace_manager)


def deserialize_rdflib_term(x):
    from rdflib.term import Literal
    if isinstance(x, Literal):
        x = x.toPython()
        if isinstance(x, Literal):
            x = str(x)
    return x


def triple_to_n3(trip, namespace_manager=None):
    from rdflib.term import Literal, URIRef
    p = ''
    ns = set([])
    for x in trip:
        s = serialize_rdflib_term(x, namespace_manager)
        if isinstance(x, URIRef) and s[0] != '<':
            ns.add(s.split(':', 1)[0])
        elif isinstance(x, Literal) and '^^' in s and s[-1] != '>':
            ns.add(s.split('^^', 1)[1].split(':', 1)[0])

        p += s + ' '
    return p


def triples_to_bgp(trips, namespace_manager=None, show_namespaces=False):
    # XXX: Collisions could result between the variable names of different
    # objects
    g = ""
    ns = set([])
    for y in trips:
        g += triple_to_n3(y) + ".\n"

    if (namespace_manager is not None) and show_namespaces:
        g = "".join('@prefix ' + str(x) + ': ' + y.n3() + ' .\n'
                    for x, y
                    in namespace_manager.namespaces()
                    if x in ns) + g

    return g


_none_singleton_set = frozenset([None])


def transitive_lookup(graph, start, predicate, context=None, direction=DOWN):
    res = set()
    border = set([start])
    while border:
        new_border = set()
        if direction is DOWN:
            qx = (list(border), predicate, None)
            idx = 2
        else:
            qx = (None, predicate, list(border))
            idx = 0
        itr = graph.triples_choices(qx, context=context)
        for t in itr:
            if isinstance(t[0], tuple):
                o = t[0][idx]
            else:
                o = t[idx]
            if o not in res:
                new_border.add(o)
        res |= border
        border = new_border
    res -= _none_singleton_set
    return res


transitive_subjects = transitive_lookup



import logging
import sys
import six

from .variable import Variable
from .graphObject import GraphObjectQuerier
from .propertyMixins import (
    DatatypePropertyMixin,
    ObjectPropertyMixin,
    UnionPropertyMixin)
from .rangedObjects import InRange
from .yProperty import Property
from .propertyValue import PropertyValue
from .mappedProperty import MappedPropertyClass
from .deprecation import deprecated
from random import randint
from lazy_object_proxy import Proxy


L = logging.getLogger(__name__)

__all__ = [
    "SimpleProperty",
    "DatatypeProperty",
    "ObjectProperty",
    "NoRelationshipException"]


class NoRelationshipException(Exception):

    """ Indicates that a Relationship was asked for but one could not be given. """


class SimpleProperty(six.with_metaclass(MappedPropertyClass, Property)):

    """ A property that has one or more links to literals or DataObjects """

    def __init__(self, **kwargs):
        # The 'linkName' must be made up from the class name if one isn't set
        # before initialization (typically in mapper._create_property)
        super(SimpleProperty, self).__init__(**kwargs)
        if not hasattr(self, 'linkName'):
            self.__class__.linkName = self.__class__.__name__ + "property"

        #
        # 'v' holds values that have been set on this SimpleProperty. It acts
        # as a sort of staging area before saving the values to the graph.
        self._v = []

    def has_value(self):
        """ Returns true if the :meth:`set` has been called previously """
        return len(self._v) > 0

    @deprecated('Please use has_value instead.')
    def hasValue(self):
        """ Returns true if the :meth:`set` has been called previously """
        return self.has_value()

    def has_defined_value(self):
        """
        Returns true if this property has a defined value
        """
        for x in self._v:
            if x.defined:
                return True
        return False

    def _get(self):
        """ Get values from a generator """
        for x in self._v:
            yield x

    @property
    def values(self):
        """ Get all values """
        return self._v

    @property
    def defined_values(self):
        """ Get values which are have their defined property set to True """
        return tuple(x for x in self._v if x.defined)

    @deprecated('Please use set instead.')
    def setValue(self, v):
        self.set(v)

    def get(self):
        """ If the ``Property`` has had ``load`` or ``set`` called previously, returns
        the resulting values. Also queries the configured rdf graph for values
        which are set for the ``Property``'s owner.
        """
        v = Variable("var" + str(id(self)))
        self.set(v)
        results = GraphObjectQuerier(v, self.rdf)()
        self.unset(v)
        return results

    def unset(self, v):

        for idx, val in enumerate(self._v):
            if not hasattr(v, 'idl') and isinstance(val, PropertyValue):
                deval = self.resolver.deserializer(val.identifier)
            else:
                deval = val
            if deval == v:
                break
        else: # no break
            raise Exception("Can't find value {}".format(v))
        actual_val = self._v[idx]
        actual_val.owner_properties.remove(self)
        self._v.remove(actual_val)

    def set(self, v):
        if isinstance(v, Rel):
            v = v.rel()

        if isinstance(v, InRange):
            vcname = 'Variable' + v.__class__.__name__
            vclass = v.__class__
            v.__class__ = type(vcname, (vclass, Variable), {})
            Variable.__init__(v, '_' + hex(randint(0, sys.maxsize))[2:])

        if not hasattr(v, "idl"):
            v = PropertyValue(v)

        if not self.multiple:
            self.clear()

        self._insert_value(v)

        return RelationshipProxy(Rel(self.owner, self, v))

    def _remove_value(self, v):
        assert self in v.owner_properties
        v.owner_properties.remove(self)
        self._v.remove(v)

    def clear(self):
        for x in self._v:
            self._remove_value(x)

    def _insert_value(self, v):
        self._v.append(v)
        if self not in v.owner_properties:
            v.owner_properties.append(self)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and (self.link == other.link)

    def __hash__(self):
        return hash(self.link)

    def __str__(self):
        return "{}({})".format(self.linkName,
                               " ".join(repr(x) for x in self._v))

    def __repr__(self):
        return str(self)


class DatatypeProperty(DatatypePropertyMixin, SimpleProperty):
    pass


class ObjectProperty(ObjectPropertyMixin, SimpleProperty):
    pass


class UnionProperty(UnionPropertyMixin, SimpleProperty):

    """ A Property that can handle either DataObjects or basic types """


class RelationshipProxy(Proxy):
    def __repr__(self):
        return repr(self.__wrapped__)

    def in_context(self, context):
        rel = self.__factory__
        rel.p.context = context
        return self

    def unwrapped(self):
        return self.__wrapped__


class Rel(tuple):

    """ A container for a relationship-assignment """
    _map = dict(s=0, p=1, o=2)

    def __new__(cls, s, p, o):
        return super(Rel, cls).__new__(cls, (s, p, o))

    def __getattr__(self, n):
        return self[Rel._map[n]]

    def __call__(self):
        return self.rel()

    def rel(self):
        from .relationship import Relationship
        return Relationship(
            s=(self.s if self.s.defined else None),
            p=self.p.rdf_object,
            o=(self.o if self.o.defined else None))


__yarom_mapped_classes__ = (SimpleProperty, ObjectProperty, DatatypeProperty, UnionProperty)



from .mapper import FCN
from .rdfUtils import DOWN


class ZeroOrMore(object):
    def __init__(self, identifier, predicate, direction=DOWN):
        self.identifier = identifier
        self.predicate = predicate
        self.direction = direction

    def __repr__(self):
        return "{}({}, {}, {})".format(FCN(type(self)),
                                       repr(self.identifier),
                                       repr(self.predicate),
                                       repr(self.direction))



from __future__ import print_function
import importlib as IM
import logging
import rdflib as R
import yarom
from itertools import count
from .rdfTypeResolver import RDFTypeResolver
from six import with_metaclass
from .mapperUtils import parents_str
from .utils import FCN
from pprint import pformat


__all__ = ["Mapper",
           "FCN",
           "UnmappedClassException"]

L = logging.getLogger(__name__)


class UnmappedClassException(Exception):
    pass


class MapperMeta(type):
    def __call__(self, *args, **kwargs):
        x = super(MapperMeta, self).__call__(*args, **kwargs)
        return x


class Mapper(with_metaclass(MapperMeta, object)):
    _instances = dict()

    @classmethod
    def get_instance(cls, *args):
        if args not in cls._instances:
            cls._instances[args] = Mapper(*args)

        return cls._instances[args]

    def __init__(self, base_class_names, base_namespace=None, imported=()):

        """ Maps class names to classes """
        self.MappedClasses = dict()

        """ Maps classes to decorated versions of the class """
        self.DecoratedMappedClasses = dict()

        """ Maps class names to parents of the related class """
        self.DataObjectsParents = dict()

        """ Maps class names to children of the related class """
        self.DataObjectsChildren = dict()

        """ Maps RDF types to properties of the related class """
        self.RDFTypeTable = dict()

        if not isinstance(base_class_names, (tuple, list)):
            raise Exception('base_class_names argument must be either a tuple'
                            ' or list')
        """ Names for the base classes """
        self.base_class_names = tuple(base_class_names)
        self.base_modules = set(n.rsplit('.', 1)[0] for n in base_class_names)

        """ The base class for objects that will be mapped.

        Defined once the module containing the class is loaded
        """
        self.base_classes = dict()

        if base_namespace is None:
            base_namespace = R.Namespace("http://example.com#")
        elif not isinstance(base_namespace, R.Namespace):
            base_namespace = R.Namespace(base_namespace)

        """ Base namespace used if a mapped class doesn't define its own """
        self.base_namespace = base_namespace

        """ Resolver can only be resolved once the base class is found """
        self.resolver = None
        self.mapdepth = 0

        """ Modules that have already been loaded """
        self.modules = dict()

        """ Module relationships """
        self.ModuleDependencies = dict()
        self.ModuleDependents = dict()

        self.imported_mappers = imported

        self.loading_module = None
        self.class_ordering = dict()

    def decorate_class(self, cls):
        return cls

    def add_class(self, cls):
        cname = FCN(cls)
        maybe_cls = self._lookup_class(cname)
        if maybe_cls is not None:
            if maybe_cls is cls:
                return False
            else:
                if hasattr(maybe_cls, 'on_mapper_remove_class'):
                    maybe_cls.on_mapper_remove_class(self)
        L.debug("Adding class %s@0x%x", cls, id(cls))

        self.MappedClasses[cname] = cls
        self.DecoratedMappedClasses[cls] = self.decorate_class(cls)
        parents = cls.__bases__
        L.debug('parents %s', parents_str(cls))
        # cls is the child to mapped parents
        mapped_parents = tuple(x for x in parents
                               if self._lookup_class(FCN(x)) is x)
        self.DataObjectsParents[cname] = mapped_parents

        for parent in mapped_parents:
            pname = FCN(parent)
            sibs = self.DataObjectsChildren.get(pname, set([]))
            sibs.add(cname)
            self.DataObjectsChildren[pname] = sibs

        for child_name, parents in self.DataObjectsParents.items():
            child = self.MappedClasses[child_name]
            if cls in child.__bases__ and cls not in parents:
                # cls is the parent to a mapped child
                self.DataObjectsParents[child_name] = (cls,) + parents
                children = self.DataObjectsChildren.get(cname, set([]))
                children.add(child_name)
                self.DataObjectsChildren[cname] = children

        if hasattr(cls, 'on_mapper_add_class'):
            cls.on_mapper_add_class(self)

        # This part happens after the on_mapper_add_class has run since the
        # class has an opportunity to set its RDF type based on what we provide
        # in the Mapper.
        self.RDFTypeTable[cls.rdf_type] = cls
        self.class_ordering = self._compute_class_ordering()
        return True

    def remap(self):
        """ Calls `map` on all of the registered classes """
        classes = set(self.MappedClasses.values())
        ordering_map = self.class_ordering

        sorted_classes = sorted(classes,
                                key=lambda y: ordering_map[FCN(y)])

        for x in sorted_classes:
            x.map()

    def _compute_class_ordering(self):
        res = dict()
        base_names = self.base_class_names

        def helper(n, gen):
            res[n] = next(gen)
            children = self.DataObjectsChildren.get(n, ())
            for c in children:
                helper(c, gen)
        it = count(0)
        for base in base_names:
            helper(base, it)

        return res

    def unmap_all(self):
        for cls in self.MappedClasses:
            cls.unmap()

    def deregister_all(self):
        keys = self.base_class_names
        for cname in keys:
            L.debug("Removing class %s", cname)
            maybe_base_class = self.MappedClasses.get(cname, None)
            if maybe_base_class is not None:
                self.remove_class(maybe_base_class)
                del self.base_classes[cname]
        L.debug('After deregister_all %s', self)

    def remove_class(self, cls):
        L.debug("Removing class %s", cls)
        if hasattr(cls, 'on_mapper_remove_class'):
            cls.on_mapper_remove_class(self)

        cname = FCN(cls)

        parents = self.DataObjectsParents[cname]

        if cname in self.DataObjectsParents:
            del self.DataObjectsParents[cname]

        if cname in self.DataObjectsChildren:
            children = tuple(self.DataObjectsChildren[cname])
            for child in children:
                self.remove_class(self.MappedClasses[child])
            del self.DataObjectsChildren[cname]

        for parent in parents:
            pname = FCN(parent)
            sibs = self.DataObjectsChildren[pname]
            sibs.remove(cname)

        if hasattr(cls, 'rdf_type') and cls.rdf_type in self.RDFTypeTable:
            del self.RDFTypeTable[cls.rdf_type]
        if cname in self.MappedClasses:
            del self.MappedClasses[cname]
        self.class_ordering = self._compute_class_ordering()

    def resolve_class(self, uri):
        cr = self.load_module('yarom.classRegistry')
        # look up the class in the registryCache
        c = self.RDFTypeTable.get(uri)
        if c is not None:
            # if it is in the regCache, then return the class;
            return c
        else:
            # otherwise, attempt to load into the cache by
            # reading the RDF graph.
            re = cr.RegistryEntry()
            re.rdfClass(uri)
            for cd in self.get_class_descriptions(re):
                # TODO: if load fails, attempt to construct the class
                cls = self.load_class_from_description(cd)
                cls.map()
                return cls

    def load_class_from_description(self, cd):
        # TODO: Undo the effects to YAROM of loading a class when the
        #       module doesn't, in fact, have the class being searched
        #       for.
        mod_name = cd.moduleName.one()
        class_name = cd.className.one()
        mod = self.load_module(mod_name)
        if mod is not None:
            if hasattr(mod, class_name):
                cls = getattr(mod, class_name)
                if cls is not None:
                    return cls
            else:
                raise Exception("Cannot find class " + class_name)
        else:
            # TODO: Load the module from the graph
            raise Exception("Cannot find module " + class_name)

    def get_class_descriptions(self, re):
        mod_data = []
        for x in re.load():
            for y in x.pythonClass():
                mod_data.append(y)
        return mod_data

    def load_module(self, module_name):
        """ Loads the module. """
        module = self.lookup_module(module_name)
        if not module:
            module = IM.import_module(module_name)
            return self.process_module(module)
        else:
            return module

    def process_module(self, module=None, cb=None, module_name=None):
        if module is None:
            if cb is None:
                raise ValueError("At least one of cb or module argument must"
                                 " be provided")
            if module_name is None:
                raise ValueError("At least one of module argument or"
                                 " module_name must be provided")
        if module_name is None:
            module_name = module.__name__

        L.log(5, "%sLOADING %s", ' ' * self.mapdepth, module_name)
        self.mapdepth += 1
        old_mapper = yarom.MAPPER
        yarom.MAPPER = self

        self.add_module_dependency(module_name)
        previously_loading = self.loading_module
        self.loading_module = module_name

        if cb is not None:
            x = cb()
            if module is None:
                module = x

        self.modules[module_name] = module
        for c in self._module_load_helper(module):
            if hasattr(c, 'after_mapper_module_load'):
                c.after_mapper_module_load(self)

        self.loading_module = previously_loading
        self.mapdepth -= 1
        yarom.MAPPER = old_mapper
        return module

    def lookup_module(self, module_name):
        m = self.modules.get(module_name, None)
        if m is None:
            for p in self.imported_mappers:
                m = p.lookup_module(module_name)
                if m:
                    break
        return m

    def load_class(self, cname_or_mname, cnames=None):
        if cnames:
            mpart = cname_or_mname
        else:
            mpart, cpart = cname_or_mname.rsplit('.', 1)
            cnames = (cpart,)
        m = self.load_module(mpart)
        try:
            res = tuple(self.DecoratedMappedClasses[c]
                        if c in self.DecoratedMappedClasses
                        else c
                        for c in
                        (getattr(m, cname) for cname in cnames))

            return res[0] if len(res) == 1 else res
        except AttributeError:
            raise UnmappedClassException(cnames)

    def add_module_dependency(self, module_name):
        if self.loading_module:
            depends = self.ModuleDependencies.get(self.loading_module, set())
            depends.add(module_name)
            self.ModuleDependencies[self.loading_module] = depends

            callers = self.ModuleDependents.get(module_name, set())
            callers.add(self.loading_module)
            self.ModuleDependents[module_name] = callers

    def _module_load_helper(self, module):
        # TODO: Make this class selector pluggable and decouple the Resolver
        # init from this -- maybe put it in a callback of some kind
        return self.handle_mapped_classes(getattr(module, '__yarom_mapped_classes__', ()))

    def handle_mapped_classes(self, classes):
        res = []
        for cls in classes:
            # This previously used the
            full_class_name = FCN(cls)
            if full_class_name in self.base_class_names:
                L.debug('Setting base class %s', full_class_name)
                self.base_classes[full_class_name] = cls
            if isinstance(cls, type) and self.add_class(cls):
                res.append(cls)
                if self.base_class_names[0] == full_class_name:
                    self.resolver = Resolver(self)
        return res

    def _merged_base_classes(self):
        ret = tuple(self.base_classes.values())
        for p in self.imported_mappers:
            ret += p._merged_base_classes()
        return ret

    def oid(self, identifier_or_rdf_type, rdf_type=False):
        """ Create an object from its rdf type

        Parameters
        ----------
        identifier_or_rdf_type : :class:`str` or :class:`rdflib.term.URIRef`
            If `rdf_type` is provided, then this value is used as the id
            for the newly created object. Otherwise, this value will be the
            :attr:`rdf_type` of the object used to determine the Python type
            and the object's identifier will be randomly generated.
        rdf_type : :class:`str`, :class:`rdflib.term.URIRef`, :const:`False`
            If provided, this will be the :attr:`rdf_type` of the newly created
            object.

        Returns
        -------
           The newly created object

        """
        L.debug("%s identifier or rdf type", identifier_or_rdf_type)
        identifier = identifier_or_rdf_type
        if not rdf_type:
            rdf_type = identifier_or_rdf_type
            identifier = False

        L.debug("oid making a {} with ident {}".format(rdf_type, identifier))
        c = self.RDFTypeTable[rdf_type]
        # if its our class name, then make our own object
        # if there's a part after that, that's the property name
        o = None
        if identifier:
            o = c(ident=identifier)
        else:
            o = c(generate_key=True)
        return o

    def compare_types(self, a, b):
        try:
            ordr = self.class_ordering
            return ordr[FCN(a)] - ordr[FCN(b)]
        except KeyError as e:
            raise Exception(
                "The given type is not in the class ordering: " +
                str(e))

    def get_most_specific_rdf_type(self, types):
        """ Gets the most specific rdf_type.

        Returns the URI corresponding to the lowest in the class hierarchy from
        among the given URIs.
        """
        # TODO: Use the MappedClasses and DataObjectsParents to get the root
        least = self.base_classes[self.base_class_names[0]]
        for x in types:
            try:
                xtype = self.RDFTypeTable[x]
                if self.compare_types(xtype, least) > 0:
                    least = self.RDFTypeTable[x]
            except KeyError:
                pass
        return least.rdf_type

    def lookup_class(self, cname):
        """ Gets the class corresponding to a fully-qualified class name """
        ret = self._lookup_class(cname)
        if ret is None:
            raise UnmappedClassException((cname,))
        return ret

    def _lookup_class(self, cname):
        c = self.MappedClasses.get(cname, None)
        if c is None:
            for p in self.imported_mappers:
                c = p._lookup_class(cname)
                if c:
                    break
        else:
            L.debug('%s.lookup_class("%s") %s@%s',
                    FCN(type(self)), cname, c, hex(id(c)))
        return c

    def mapped_classes(self):
        for p in self.imported_mappers:
            for c in p.mapped_classes():
                yield
        for c in self.MappedClasses.values():
            yield c

    def resolver(self):
        return self.resolver

    def __str__(self):
        return "Mapper[" + \
            pformat({x: getattr(self, x) for x in ('MappedClasses',
                                                   'DataObjectsParents',
                                                   'DataObjectsChildren',
                                                   'RDFTypeTable',
                                                   'ModuleDependencies')}) + \
            "]"


class _ClassOrderable(object):
    def __init__(self, cls):
        self.cls = cls

    def __eq__(self, other):
        self.cls is other.cls

    def __gt__(self, other):
        res = False
        ocls = other.cls
        scls = self.cls
        if issubclass(ocls, scls) and not issubclass(scls, ocls):
            res = True
        elif issubclass(scls, ocls) == issubclass(ocls, scls):
            res = scls.__name__ > ocls.__name__
        return res

    def __lt__(self, other):
        res = False
        ocls = other.cls
        scls = self.cls
        if issubclass(scls, ocls) and not issubclass(ocls, scls):
            res = True
        elif issubclass(scls, ocls) == issubclass(ocls, scls):
            res = scls.__name__ < ocls.__name__
        return res


class Resolver(RDFTypeResolver):
    instances = dict()

    @classmethod
    def get_instance(cls, mapper):
        if mapper not in cls.instances:
            # XXX: This is a bit of a kludge. YAROM privileges one of the base
            # classes which is the DataObject.
            cls.instances[mapper] = Resolver(mapper)
        return cls.instances[mapper]

    def __init__(self, mapper):
        from .rdfUtils import deserialize_rdflib_term
        DataObject = mapper.base_classes[mapper.base_class_names[0]]
        super(Resolver, self).__init__(DataObject.rdf_type,
                                       mapper.get_most_specific_rdf_type,
                                       mapper.oid,
                                       deserialize_rdflib_term)
        self.mapper = mapper



from .dataUser import DataUser

__all__ = ["Property"]


class Property(DataUser):

    """ Store a value associated with a DataObject

    Properties can be be accessed like methods. A method call like::

        a.P()

    for a property ``P`` will return values appropriate to that property for
    ``a``, the `owner` of the property.

    Parameters
    ----------
    owner : yarom.dataObject.DataObject
        The owner of this property
    name : string
        The name of this property. Can be accessed as an attribute like::

            owner.name

    """

    # Indicates whether the Property is multi-valued
    multiple = False
    link = None
    linkName = None

    def __init__(self, owner=False, **kwargs):
        super(Property, self).__init__(**kwargs)
        self.owner = owner

    def get(self, *args):
        """
        Get the things which are on the other side of this property

        The return value must be iterable. For a ``get`` that just returns
        a single value, an easy way to make an iterable is to wrap the
        value in a tuple like ``(value,)``.

        Derived classes must override.
        """

        raise NotImplementedError()

    def set(self, *args, **kwargs):
        """
        Set the value of this property

        Derived classes must override.
        """

        raise NotImplementedError()

    def one(self):
        """
        Returns a single value for the ``Property`` whether or not it is
        multivalued.
        """

        try:
            r = self.get()
            return next(iter(r))
        except StopIteration:
            return None

    def hasValue(self):
        """
        Returns true if the Property has any values set on it.

        This may be defined differently for each property
        """
        return False

    @property
    def values(self):
        raise NotImplementedError()

    def __call__(self, *args, **kwargs):
        """
        If arguments are passed to the ``Property``, its ``set`` method is
        called. Otherwise, the ``get`` method is called. If the ``multiple``
        member for the ``Property`` is set to ``True``, then a Python set
        containing the associated values is returned. Otherwise, a single bare
        value is returned.
        """

        if len(args) > 0 or len(kwargs) > 0:
            return self.set(*args, **kwargs)
        else:
            r = self.get(*args, **kwargs)
            if self.multiple:
                return set(r)
            else:
                try:
                    return next(iter(r))
                except StopIteration:
                    return None



import rdflib as R
import logging
import hashlib
import six
import random

from yarom import yarom_import
from .mappedClass import MappedClass
from .mappedProperty import MappedPropertyClass
from .dataUser import DataUser
from .configure import BadConf
from .rdfUtils import triples_to_bgp
from .graphObject import (
    GraphObject,
    GraphObjectQuerier,
    ComponentTripler,
    HeroTripler,
    ReferenceTripler,
    DescendantTripler,
    IdentifierMissingException)

SimpleProperty, DatatypeProperty, ObjectProperty = \
        yarom_import('yarom.simpleProperty', ('SimpleProperty',
                                              'DatatypeProperty',
                                              'ObjectProperty'))

"""
.. autoclass:: DataObject
"""

# NOTE: This module (and all modules containing DataObject sub-types) may
# be reloaded. Objects should not be created in the top-level of this
# module.

L = logging.getLogger(__name__)


def _bnode_to_var(x):
    return "?" + x


def get_hash_function(method_name):
    if method_name == "sha224":
        return hashlib.sha224
    elif method_name == "md5":
        return hashlib.md5
    elif method_name in hashlib.algorithms_available:
        return (lambda data: hashlib.new(method_name, data))


class DataObject(six.with_metaclass(MappedClass, GraphObject, DataUser)):

    """
    An object backed by the database

    Attributes
    -----------
    rdf_type : rdflib.term.URIRef
        The RDF type URI for objects of this type
    rdf_namespace : rdflib.namespace.Namespace
        The rdflib namespace (prefix for URIs) for objects from this class
    properties : list of Property
        Properties belonging to this object
    owner_properties : list of Property
        Properties belonging to parents of this object
    """

    _openSet = set()
    _closedSet = set()

    configuration_variables = {
        "rdf.namespace": {
            "description": "Namespaces for DataObject sub-classes will, by "
            "default, be based off of this. For example, a subclass named A "
            "would have a namespace '[rdf.namespace]A/'",
            "type": R.Namespace,
            "directly_configureable": True},
        "dataObject.identifier_hash": {
            "description": "The hash method used for object identifiers. "
            "Defaults to md5.",
            "type": "sha224, md5, or one of the types accepted by"
            "hashlib.new()",
            "directly_configureable": True},
    }
    base_namespace = R.Namespace("http://openworm.org/entities/")

    @classmethod
    def open_set(self):
        """ The open set contains items that must be saved directly in order for
        their data to be written out
        """
        return self._openSet

    def __init__(
            self,
            ident=False,
            var=False,
            key=False,
            generate_key=False,
            **kwargs):
        """A subclass of DataObject cannot have any required positional arguments.

        Parameters
        ----------
        ident : rdflib.term.URIRef or str
            The identifier for this DataObject
        var : str
            In lieu of `ident`, sets the variable for this object
        key : str or object
            In lieu of `ident` or `var`, sets the identifier for this
            DataObject using the key value. For a namespace `ex:` and key `a`,
            the identifier would be `ex:a`.
        generate_key : bool
            If true generates a random key value
        kwargs : dict
            Values to set for named properties
        """
        try:
            super(DataObject, self).__init__()
        except BadConf as e:
            six.raise_from(
                Exception("You may need to connect to " +
                          "a database before continuing."), e)

        if not self.__class__.mapped:
            raise Exception(
                ("The class `{0}({1})` has not been mapped. You should call "
                 "`{0}.map()` before creating any instances.").format(
                    self.__class__.__name__, ", ".join(p.__name__ + '@' + hex(id(p)) for p in self.__class__.mro())))

        self._id = False

        if ident:
            if isinstance(ident, R.URIRef):
                self._id = ident
            else:
                self._id = R.URIRef(ident)
        elif var:
            self._id_variable = R.Variable(var)
        # TODO: Support a key function that generates the key based on live
        # values of the object (e.g., property values)
        elif key:
            self.setKey(key)
        elif generate_key:
            self.setKey(random.random())
        else:
            # Randomly generate an identifier if the derived class can't
            # come up with one from the start. Ensures we always have something
            # that functions as an identifier
            v = (random.random(), random.random())
            cname = self.__class__.__name__
            self._id_variable = R.Variable(cname + "_" + hashlib.md5(
                str(v).encode()).hexdigest())

        for x in self.__class__.dataObjectProperties:
            self.attachProperty(x)

        existing_property_names = [x.linkName for x in self.properties]
        for propName in kwargs.keys():
            if propName not in existing_property_names:
                raise ValueError(
                    "No such argument {} to {}::__init__".format(
                        propName,
                        self.__class__.__name__))

        for x in self.properties:
            if x.linkName in kwargs:
                self.relate(x.linkName, kwargs[x.linkName])

        if isinstance(self, PropertyDataObject):
            self.relate(
                'rdf_type_property',
                RDFProperty.getInstance(),
                RDFTypeProperty)
        elif isinstance(self, TypeDataObject):
            self.relate(
                'rdf_type_property',
                RDFSClass.getInstance(),
                RDFTypeProperty)
        elif isinstance(self, RDFProperty):
            self.relate(
                'rdf_type_property',
                RDFSClass.getInstance(),
                RDFTypeProperty)
        elif isinstance(self, RDFSClass):
            self.relate('rdf_type_property', self, RDFTypeProperty)
        else:
            self.relate(
                'rdf_type_property',
                self.rdf_type_object,
                RDFTypeProperty)

    @classmethod
    def identifier_hash_method(self, o):
        return get_hash_function(
            self.conf.get(
                'dataObject.identifier_hash',
                'md5'))(o)

    def set_property_values(self, values):
        for k in six.iterkeys(values.keys):
            o = values[k]
            if o is not None:
                setattr(self, k, o)

    def make_identifier_from_properties(self, *properties):
        if len(properties) == 0:
            raise Exception("No properties provided to make identifier")
        data = ""
        for propName in properties:
            for value in getattr(self, propName).values:
                data += value.idl.n3()
        if len(data) == 0:
            raise Exception("No properties to make identifier")
        return self.make_identifier(data)

    @property
    def defined(self):
        """Returns `True` if this object has an identifier

        To define a custom identifier, override :meth:`defined_augment` to return
        True when your custom identifier would be defined. You must also override
        :meth:`identifier_augment`
        """
        # TODO: Add a check for circular definition of "defined" status by
        #       defined_augment like:
        #
        #           a is defined if b is defined
        #           b is defined if a is defined.
        #
        #       such definitions are legal and if one or the other has its
        #       ident set explicitly, then there would be done unbounded
        #       recursion.
        if self._id:
            return True
        else:
            return self.defined_augment()

    def defined_augment(self):
        """ This fuction must return False if :meth:`identifier_augment` would
        raise an :exc:`~yarom.graphObject.IdentifierMissingException`. Override
        it when defining a non-standard identifier for subclasses of DataObjects.
        """
        return False

    def variable(self):
        """ Returns the variable to be usedin queries with this DataObject

        Raises
        ------

        IdentifierMissingException

        """
        if self._id_variable is not None:
            return self._id_variable
        else:
            raise IdentifierMissingException(self)

    def setKey(self, key):
        if isinstance(key, str):
            self._id = self.make_identifier_direct(key)
        else:
            self._id = self.make_identifier(key)

    def relate(self, linkName, other, prop=False):
        existing_property_names = [x.linkName for x in self.properties]
        if linkName in existing_property_names:
            if prop and not isinstance(getattr(self, linkName), prop):
                L.warning(
                    "A property class, {}, was provided to relate, but it"
                    " will be ignored since there's already a property, {},"
                    " with the given linkName '{}' and it has a different"
                    " class".format(prop, getattr(self, linkName), linkName))
            p = getattr(self, linkName)
        else:
            if not prop:
                # Make up a property class
                property_type = None
                if isinstance(other, DataObject):
                    property_type = ObjectProperty
                else:
                    property_type = DatatypeProperty
                prop = self._create_related_property(linkName, property_type)
            elif not hasattr(prop, 'linkName'):
                prop.linkName = linkName
            p = self.attachProperty(prop)
        return p.set(other)

    def _create_related_property(self, linkName, property_type):
        link = type(self).rdf_namespace[linkName]
        return MappedPropertyClass(
            linkName, (property_type,), dict(
                link=link, linkName=linkName, multiple=True))

    def attachProperty(self, prop):
        if not hasattr(prop, 'linkName'):
            raise Exception("The given property class cannot be attached"
                            " because it has no `linkName` attribute")
        p = prop(resolver=self.mapper.resolver, owner=self)

        if hasattr(self, prop.linkName):
            raise Exception(
                "Cannot attach property '{}'. A property must have a different \
                        name from any attributes in DataObject".format(
                    prop.linkName))
        self.properties.append(p)
        setattr(self, p.linkName, p)
        return p

    def get_defined_component(self):
        g = ComponentTripler(self, generator=True)()
        if not isinstance(g, R.Graph):
            h = R.Graph()
            for t in g:
                h.add(t)
            g = h
        g.namespace_manager = self.namespace_manager
        return g

    def __eq__(self, other):
        return ((isinstance(other, DataObject) and
                 (self.idl == other.idl)) or
                (isinstance(other, R.URIRef) and
                 (self.idl == other)))

    def __hash__(self):
        return hash(self.idl)

    def __str__(self):
        return self.namespace_manager.normalizeUri(self.idl)

    def __repr__(self):
        return self.__str__()

    @classmethod
    def add_to_open_set(cls, o):
        cls._openSet.add(o)

    @classmethod
    def remove_from_open_set(cls, o):
        if o not in cls._closedSet:
            cls._openSet.remove(o)
            cls._closedSet.add(o)

    @classmethod
    def extract_unique_part(cls, uri):
        if uri.startswith(cls.rdf_namespace):
            return uri[:len(cls.rdf_namespace)]
        else:
            raise Exception(
                "This URI ({}) doesn't start with the appropriate namespace ({})".format(
                    uri,
                    cls.rdf_namespace))

    @classmethod
    def make_identifier(cls, data):
        # NOTE: The "a" prefix allows all identifiers to nicely reduce to
        # abbreviated form in n3
        return R.URIRef(
            cls.rdf_namespace[
                "a" +
                cls.identifier_hash_method(
                    str(data).encode()).hexdigest()])

    @classmethod
    def make_identifier_direct(cls, string):
        if not isinstance(string, str):
            raise Exception("make_identifier_direct only accepts strings")
        from six.moves import urllib
        return R.URIRef(cls.rdf_namespace[urllib.parse.quote(string)])

    @property
    def identifier(self):
        """ The identifier for this object in the rdf graph.

        This identifier may be randomly generated, but an identifier returned
        from the graph can be used to retrieve the specific object that it
        refers to.

        If it is desireable to customize the identifier, a subclass of
        DataObject should override :meth:`identifier_augment` rather than this
        method.

        Returns
        -------
        :class:`rdflib.term.URIRef`
        """
        if self._id:
            return self._id
        else:
            return self.identifier_augment()

    def identifier_augment(self):
        """ Override this method to define an identifier in lieu of one explicity set.

        One must also override :meth:`defined_augment` to return True whenever
        this method could return a valid identifier.
        :exc:`~yarom.graphObject.IdentifierMissingException` should be
        raised if an identifier cannot be generated by this method.

        Raises
        ------

        IdentifierMissingException

        """
        raise IdentifierMissingException(self)

    def triples(self):
        """ Returns 3-tuples of the connected component of the object graph
        starting from this object.

        Returns
        --------
        An iterable of triples
        """
        return self.get_defined_component()

    def graph_pattern(self, shorten=False):
        """ Get the graph pattern for this object.

        It should be as simple as converting the result of triples() into a BGP

        Parameters
        ----------
        query : bool
            Indicates whether or not the graph_pattern is to be used for
            querying (as in a SPARQL query) or for storage
        shorten : bool
            Indicates whether to shorten the URLs with the namespace manager
            attached to the ``self``
        """

        nm = None
        if shorten:
            nm = self.namespace_manager
        return triples_to_bgp(
            self.get_defined_component(),
            namespace_manager=nm)

    def load(self):
        for ident in GraphObjectQuerier(self, self.rdf)():
            types = set()
            for rdf_type in self.rdf.objects(ident, R.RDF['type']):
                types.add(rdf_type)
            the_type = self.mapper.get_most_specific_rdf_type(types)
            yield self.mapper.oid(ident, the_type)

    def resolve(self):
        """ Resolve this object from the graph.

        Gets the value for each of the properties attached to this object. In
        contrast to :meth:`load`, this method requires that the object already
        has its identifier.
        """

        for p in self.properties:
            values = set(p.get())
            for v in values:
                p.set(v)

    def save(self):
        """ Write in-memory data to the database. Derived classes should call this to update
        the store.

        Dual to retract.
        """
        self.add_statements(self.get_defined_component())

    def retract(self):
        """ Remove this object from the data store.

        Retract removes an object and everything it points to, transitively, and everything
        which points to it.

        Dual to save.
        """
        self.retract_statements(self.get_defined_component())

    def save_object(self):
        """ Write in-memory data to the database. Derived classes should call this to update
        the store.

        Dual to retract_object.
        """
        self.add_statements(DescendantTripler(self)())

    def retract_object(self):
        """ Remove this object from the data store.

        Retract removes an object and everything it points to, transitively, and everything
        which points to it.

        Dual to save_object.
        """
        self.retract_statements(HeroTripler(self)())

    def retract_objectG(self):
        """ Remove this object from the data store.

        Retract removes an object and everything it points to, transitively, and everything
        which points to it.

        Dual to save_objectG.
        """
        g = HeroTripler(self, self.rdf)()
        self.retract_statements(g)

    def retract_references(self):
        """ Remove all references directly to or made by this object """
        self.retract_statements(ReferenceTripler(self)())

    def retract_referencesG(self):
        """ Remove all references directly to or made by this object """
        self.retract_statements(ReferenceTripler(self, self.rdf)())

    def __getitem__(self, x):
        try:
            return self.conf[x]
        except KeyError:
            raise Exception(
                "You attempted to get the value `{}' from `{}'. It isn't here."
                " Perhaps you misspelled the name of a Property?".format(
                    x, self))

    def get_owners(self, property_name):
        """ Return the owners along a property pointing to this object """
        res = []
        for x in self.owner_properties:
            if isinstance(x, SimpleProperty):
                if str(x.linkName) == str(property_name):
                    res.append(x.owner)
        return res


def validateG(do_type):
    """ Given a DataObject type, call validate() on all objects of that type in the Python object graph """


def validate(do_type):
    """ Given a DataObject type, call validate() on all objects of that type in the RDF object graph """


class TypeDataObject(DataObject):
    pass


class DataObjectSingleton(DataObject):
    instance = None

    def __init__(self, *args, **kwargs):
        if type(self)._gettingInstance:
            super(DataObjectSingleton, self).__init__(*args, **kwargs)
        else:
            raise Exception(
                "You must call getInstance to get " +
                type(self).__name__)

    @classmethod
    def getInstance(cls):
        if cls.instance is None:
            cls._gettingInstance = True
            cls.instance = cls()
            cls._gettingInstance = False

        return cls.instance


class RDFSClass(DataObjectSingleton):

    """ The DataObject corresponding to rdfs:Class """
    # XXX: This class may be changed from a singleton later to facilitate dumping
    #      and reloading the object graph
    rdf_type = R.RDFS['Class']
    auto_mapped = True

    def __init__(self):
        super(RDFSClass, self).__init__(R.RDFS["Class"])


class RDFProperty(DataObjectSingleton):

    """ The DataObject corresponding to rdf:Property """
    rdf_type = R.RDF['Property']

    def __init__(self):
        super(RDFProperty, self).__init__(R.RDF["Property"])


class RDFTypeProperty(ObjectProperty):
    link = R.RDF['type']
    linkName = "rdf_type_property"
    owner_type = DataObject
    value_type = RDFSClass
    multiple = True


class RDFSSubClassOfProperty(ObjectProperty):
    link = R.RDFS['subClassOf']
    linkName = "rdfs_subClassOf"
    owner_type = RDFSClass
    value_type = RDFSClass
    multiple = True


class PropertyDataObject(DataObject):

    """ A PropertyDataObject represents the property-as-object.

    Try not to confuse this with the Property class
    """


class RDFSDomainProperty(ObjectProperty):
    link = R.RDFS['domain']
    linkName = "rdfs_domain"
    owner_type = RDFProperty
    value_type = RDFSClass
    multiple = True


class RDFSRangeProperty(ObjectProperty):
    link = R.RDFS['range']
    linkName = "rdfs_range"
    owner_type = RDFProperty
    value_type = RDFSClass
    multiple = True


__yarom_mapped_classes__ = (DataObject, PropertyDataObject, RDFProperty, RDFSClass,
                            DataObjectSingleton, TypeDataObject)



from __future__ import print_function
import six
import warnings
import functools


class deprecated(object):

    def __init__(self, message_or_callable=None):
        if isinstance(message_or_callable, six.string_types):
            self.message = message_or_callable
            self.func = None
        else:
            self.func = message_or_callable
            self.message = None

    def __call__(self, *args, **kwargs):
        '''This is a decorator which can be used to mark functions
        as deprecated. It will result in a warning being emitted
        when the function is used.'''
        if self.message:
            func = args[0]

            @functools.wraps(func)
            def new_func(*args, **kwargs):
                self.warn(func, self.message)
                return func(*args, **kwargs)
            return new_func
        else:
            self.warn(self.func)
            return self.func(*args, **kwargs)

    def warn(self, func, message=None):
        warnings.warn_explicit(
                    "Call to deprecated function {}{}".format(
                        func.__name__, "."
                        if message is None else ": " + str(message)),
                    category=DeprecationWarning,
                    filename=func.func_code.co_filename,
                    lineno=func.func_code.co_firstlineno + 1)


if __name__ == '__main__':
    @deprecated("The MESSAGE")
    def my_func():
        print("yelp")

    @deprecated
    def mu_func():
        print("yolp")

    my_func()
    mu_func()



from yarom import yarom_import

DataObject = yarom_import('yarom.dataObject.DataObject')


class Relationship(DataObject):
    """ A Relationship is typically related to a property and is an object that
        one points to for talking about the property relationship.

        For SimpleProperty objects, this acts like a RDF Reified triple.
        """
    _ = [{'name': 'subject', 'multiple': False},
         {'name': 'property', 'multiple': False},
         {'name': 'object', 'multiple': False}]

    def _ident_data(self):
        return [self.subject.values,
                self.property.values,
                self.object.values]

    def defined_augment(self):
        for p in self._ident_data():
            if len(p) < 1:
                return False
        return True

    def identifier_augment(self):
        x = self.make_identifier_from_properties('subject',
                                                 'property',
                                                 'object')
        return x




__all__ = [
    'warn_mismapping',
    'log_raise_mismapping_exception',
    'raise_mismapping_exception',
    'warn_mismapping',
    'parents_str']


def mismapMessage(mapping, key, should_be, is_actually):
    return "Mismapping of {} in {}. Is {}. Should be {}".format(
        key,
        mapping,
        is_actually,
        should_be)


def warn_mismapping(logger, mapping, key, should_be, is_actually=None):
    logger.warning(create_mismap_message(mapping, key, should_be, is_actually))


def raise_mismapping_exception(mapping, key, should_be, is_actually=None):
    raise Exception(create_mismap_message(mapping, key, should_be, is_actually))


def log_raise_mismapping_exception(
        logger,
        mapping,
        key,
        should_be,
        is_actually=None):
    warning_message = create_mismap_message(mapping, key, should_be, is_actually)
    logger.warning(warning_message)
    raise Exception(warning_message)


def create_mismap_message(mapping, key, should_be, is_actually):
    is_actually = get_actual(mapping, key, is_actually)
    return mismapMessage(mapping, key, should_be, is_actually)


def get_actual(mapping, key, is_actually):
    if is_actually is None:
        is_actually = mapping[key]
    return is_actually


def parents_str(cls):
    return ", ".join(p.__name__ + '@' + hex(id(p)) for p in cls.mro())



from rdflib import ConjunctiveGraph
import os
import transaction
import traceback
import logging

from .data import RDFSource

L = logging.getLogger(__name__)


class ZODBSource(RDFSource):

    """ Reads from and queries against a configured Zope Object Database.

        If the configured database does not exist, it is created.

        The database store is configured with::

            "rdf.source" = "ZODB"
            "rdf.store_conf" = <location of your ZODB database>

        Leaving unconfigured simply gives an in-memory data store.
    """
    name = "zodb"

    def __init__(self, *args, **kwargs):
        super(ZODBSource, self).__init__(*args, **kwargs)
        self.conf['rdf.store'] = "ZODB"

    def open(self):
        import ZODB
        from ZODB.FileStorage import FileStorage
        self.path = self.conf['rdf.store_conf']
        openstr = os.path.abspath(self.path)
        try:
            fs = FileStorage(openstr)
            self.zdb = ZODB.DB(fs)
            self.conn = self.zdb.open()
            root = self.conn.root()
            if 'rdflib' not in root:
                root['rdflib'] = ConjunctiveGraph('ZODB')
            self.graph = root['rdflib']
            try:
                transaction.commit()
            except Exception:
                # catch commit exception and close db.
                # otherwise db would stay open and follow up tests
                # will detect the db in error state
                L.warning('Forced to abort transaction on ZODB store opening')
                traceback.print_exc()
                transaction.abort()
            transaction.begin()
            self.graph.open(self.path)
        except Exception:
            transaction.abort()
            raise Exception(
                "ZODB format error. This may be a result of using two"
                " different versions of ZODB, such as between Python 3.x and"
                " Python 2.x")

    def close(self):
        if self.graph is None:
            return

        self.graph.close()

        try:
            transaction.commit()
        except Exception:
            # catch commit exception and close db.
            # otherwise db would stay open and follow up tests
            # will detect the db in error state
            traceback.print_exc()
            L.warning('Forced to abort transaction on ZODB store closing')
            transaction.abort()
        self.conn.close()
        self.zdb.close()
        self.graph = None



import logging

from rdflib import ConjunctiveGraph
from .data import RDFSource

L = logging.getLogger(__name__)


class SleepyCatSource(RDFSource):

    """ Reads from and queries against a local Sleepycat database

        The database can be configured like::

            "rdf.source" = "Sleepycat"
            "rdf.store_conf" = <your database location here>
    """
    name = 'sleepycat'

    def open(self):
        # XXX: If we have a source that's read only, should we need to set the
        # store separately??
        g0 = ConjunctiveGraph('Sleepycat')
        self.conf['rdf.store'] = 'Sleepycat'
        g0.open(self.conf['rdf.store_conf'], create=True)
        self.graph = g0
        L.debug("Opened SleepyCatSource")



# -*- coding: utf-8 -*-

from setuptools import setup
import os
import sys
from glob import glob

REQUIRED = []
with open('requirements.txt') as f:
    REQUIRED = f.read().splitlines()
    on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
    if on_rtd:
        REQUIRED.append("numpydoc")
FEATURE_DEPS = {}

PY2 = sys.version_info.major == 2

for feature_file in glob("*.requirements.txt"):
    feature, _ = feature_file.split(".", 1)

    with open(feature_file) as f:
        FEATURE_DEPS[feature] = f.read().splitlines()

VERSION = open("version.txt").read().strip()
LONG_DESCRIPTION = open("README.rst").read()

setup(
    name="YAROM",
    setup_requires=['six>=1.9.0'],
    tests_require=[] + (['mock==2.0.0'] if PY2 else []),
    install_requires=REQUIRED,
    dependency_links=[
        "git://github.com/zopefoundation/ZODB.git#egg=ZODB",
        "git://github.com/RDFLib/rdflib-zodb.git#egg=rdflib_zodb-1.0.0",
    ],
    version=VERSION,
    packages=['yarom'],
    package_data={"yarom": ['default.conf', 'rules.n3']},
    author="Mark Watts",
    author_email="wattsmark2015@gmail.com",
    description="Yet Another RDF-Object Mapper",
    long_description=LONG_DESCRIPTION,
    license="BSD 3-clause",
    url="http://yarom.readthedocs.org/en/latest/",
    download_url='https://github.com/mwatts15/YAROM/archive/v'+VERSION+'.tar.gz',
    classifiers=[
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Topic :: Scientific/Engineering'],
    extras_require=FEATURE_DEPS,
    zip_safe=False
)



#
# a class for modules that need outside objects to parameterize their behavior
# Modules inherit from this class and use their
# self['expected_configured_property']
import traceback


class ConfigValue(object):

    """ A value to be configured.

    Elements of a :class:`Configuration` are, in fact, :class:`ConfigValue` objects. They
    can be resolved an arbitrary time after the :class:`Configuration` object is created
    by calling :meth:`get`.
    """

    def get(self):
        """ Override this method to return a value when a configuration variable is accessed"""
        raise NotImplementedError


class _C(ConfigValue):

    """ A helper class that simply stores a value and can report it back with the get method.
        Subclasses ConfigValue and implements the get method.
    """

    def __init__(self, v):
        self.v = v

    def get(self):
        return self.v

    def __str__(self):
        return str(self.v)

    def __repr__(self):
        return repr(self.v)


class BadConf(Exception):

    """ Special exception subclass for alerting the user to a bad configuration
    """
    pass


class _link(ConfigValue):

    """ Helper class that groups values within a Configuration
    """

    def __init__(self, members, c):
        self.members = members
        self.conf = c

    def get(self):
        return self.conf[self.members[0]]


class Configuration(object):

    """ A simple configuration object.  Enables setting and getting key-value pairs"""
    # conf: is a configure instance to base this one on
    # dependencies are required for this class to be initialized (TODO)

    def __init__(self, **kwargs):
        for x in kwargs:
            if not isinstance(kwargs[x], ConfigValue):
                kwargs[x] = _C(kwargs[x])
        self._properties = kwargs

    def __setitem__(self, pname, value):
        if not isinstance(value, ConfigValue):
            value = _C(value)
        if (pname in self._properties) and isinstance(
                self._properties[pname], _link):
            for x in self._properties[pname].members:
                self._properties[x] = value
        else:
            self._properties[pname] = value

    def __getitem__(self, pname):
        return self._properties[pname].get()

    def __iter__(self):
        return iter(self._properties)

    def link(self, *names):
        """ Call link() with the names of configuration values that should
        always be the same to link them together
        """
        l = _link(names, self)
        for n in names:
            self._properties[n] = l

    def __contains__(self, thing):
        return (thing in self._properties)

    def __str__(self):
        return "\n".join(
            "%s = %s" %
            (k, self._properties[k]) for k in self._properties)

    def __repr__(self):
        return (
            "Configuration(**{\n" +
            ",\n".join(
                "{} : {}".format(
                    repr(k),
                    repr(
                        self._properties[k])) for k in self._properties) +
            "})")

    def __len__(self):
        return len(self._properties)

    @classmethod
    def open(cls, file_name):
        """ Open a configuration file and read it to build the internal state.

        Parameters
        ----------
        file_name : str
            The name of a configuration file encoded as JSON

        Returns
        -------
        Configuration
            a Configuration object with the configuration taken from the JSON file
        """
        import json
        with open(file_name) as f:
            c = Configuration()
            d = json.load(f)
            for k in d:
                c[k] = _C(d[k])
        c['configure.file_location'] = file_name
        return c

    def copy(self, other):
        """ Copy configuration values from a different object.

        Parameters
        ----------
        other : dict or Configuration
            A dict or Configuration object to copy the configuration from

        Returns
        -------
        self
        """
        if isinstance(other, Configuration):
            self._properties = dict(other._properties)
        elif isinstance(other, dict):
            for x in other:
                self[x] = other[x]
        return self

    def get(self, pname, default=None):
        """ Retrieve a configuration value.

        Parameters
        ----------
        pname : str
            The key of the value to return.
        default : object
            The value to return if there is no value corresponding to the given key

        Returns
        -------
        object
            The value corresponding to the key in pname or `default` if none is
            available and a default is provided.

        Raises
        ------
        KeyError
            If the given key has no associated value and no default is provided
        """
        if pname in self._properties:
            return self._properties[pname].get()
        elif (default is not None):
            return default
        else:
            traceback.print_stack()
            raise KeyError(pname)


class Configureable(object):

    """ An object which can be configured.

    A ``Configureable`` object can access a :class:`Configuration` object,
    ``Configureable.conf``, which is shared among all ``Configureable`` objects.

    The configuration variables which can affect the behavior of a class should
    be documented in the ``configuration_variables`` class variable. This table
    will be checked on each access of ``Configureable.conf``
    """

    conf = Configuration()
    """ The configuration """

    configuration_variables = dict()
    """ A table of configuration values used by the Configureable object for the
    purpose of documentation.

    The table is indexed by the configuration value. Among the data included in
    the table should be:

    - a "description" which describes how the configuration value is used
      *within the configureable object*: broad generalization about the variable
      shouldn't be here.

    - a "type" for the value of the config may also be included and may be a
      Python ``type`` or just a string description. This isn't at all intended
      to be used for type checking, but *is purely descriptive*.

    - a "directly_configureable" indicator which should be set to ``True`` if the
      value passed in to the object for configuration variable is used more-or-
      less directly by the object. Sanitization of the value or translation into
      a more specific form are acceptable for a variable that is nonetheless
      directly_configureable. On the other hand, a configuration variable that
      has its value set within the object should have directly_configureable set
      to ``False``.
    """

    def __init__(self, conf=None):
        pass

    @classmethod
    def setConf(cls, conf):
        cls.conf = conf

    def __getitem__(self, k):
        return self.conf.get(k)

    def __setitem__(self, k, v):
        self.conf[k] = v

    def get(self, pname, default=False):
        return self.conf.get(pname, default)



from yarom import yarom_import


class ObjectCollection(yarom_import('yarom.dataObject.DataObject')):

    """
    A convenience class for working with a collection of objects

    Example::

        v = ObjectCollection('unc-13 neurons and muscles')
        n = P.Neuron()
        m = P.Muscle()
        n.receptor('UNC-13')
        m.receptor('UNC-13')
        for x in n.load():
            v.value(x)
        for x in m.load():
            v.value(x)
        # Save the group for later use
        v.save()
        ...
        # get the list back
        u = ObjectCollection('unc-13 neurons and muscles')
        nm = list(u.value())


    Parameters
    ----------
    group_name : string
        A name of the group of objects

    Attributes
    ----------
    name : DatatypeProperty
        The name of the group of objects
    group_name : DataObject
        an alias for ``name``
    member : ObjectProperty
        An object in the group
    add : ObjectProperty
        an alias for ``value``

    """
    _ = ['member']
    datatypeProperties = [{'name': 'name', 'multiple': False}]

    def __init__(self, group_name=False, **kwargs):
        super(ObjectCollection, self).__init__(key=group_name, **kwargs)
        self.add = self.member
        self.group_name = self.name
        self.name(group_name)


__yarom_mapped_classes__ = (ObjectCollection,)



import logging
import rdflib as R

L = logging.getLogger(__name__)


class MappedPropertyClass(type):

    def __init__(cls, name, bases, dct):
        L.debug("INITIALIZING %s", name)
        super(MappedPropertyClass, cls).__init__(name, bases, dct)
        if 'link' in dct:
            cls.link = dct['link']

    def on_mapper_add_class(self, mapper):
        self.mapper = mapper
        # This is how we create the RDF predicate that points from the owner
        # to this property
        L.debug("REGISTERING %s", self.__name__)
        self.rdf_type = R.RDF['Property']

        return self

    def after_mapper_module_load(self, mapper):
        vt = getattr(self, 'value_type', None)
        vrt = getattr(vt, 'rdf_type', None)
        if vrt:
            self.value_rdf_type = vrt

    def map(cls):
        from .dataObject import (
            PropertyDataObject,
            RDFSDomainProperty,
            RDFSRangeProperty)

        L.debug("MAPPING %s", cls.__name__)
        if cls.link is None:
            if hasattr(cls, 'owner_type'):
                cls.link = cls.owner_type.rdf_namespace[cls.linkName]

        cls.rdf_object = PropertyDataObject(ident=cls.link)
        if hasattr(cls, 'owner_type'):
            cls.rdf_object.relate(
                'rdfs_domain',
                cls.owner_type.rdf_type_object,
                RDFSDomainProperty)

        if hasattr(cls, 'value_type'):
            cls.rdf_object.relate(
                'rdfs_range',
                cls.value_type.rdf_type_object,
                RDFSRangeProperty)

    def __lt__(self, other):
        res = False
        if issubclass(other, self) and not issubclass(self, other):
            res = True
        elif issubclass(self, other) == issubclass(other, self):
            res = self.__name__ < other.__name__
        return res



import rdflib as R

from yarom import yarom_import

DataObject = yarom_import('yarom.dataObject.DataObject')

# TODO Generate key for the RegistryEntry type as a combination of the
#      ClassDescription.classKey and rdfClass
# TODO Generate RegistryEntry objects for all mapped classes automatically
rns = R.Namespace("yarom/PythonClassRegistry/schema#")


class RegistryEntry(DataObject):

    """ A mapping from a Python class to an RDF class.

    Objects of this type are utilized in the resolution of Python classes from
    the RDF graph
    """
    rdf_type = rns["RegistryEntry"]
    rdf_namespace = R.Namespace(rdf_type + '/')
    _ = ["pythonClass", "rdfClass"]


class ClassDescription(DataObject):
    rdf_type = rns["ClassDescription"]
    rdf_namespace = R.Namespace(rdf_type + '/')
    _ = ["classKey", "className", "moduleName", "moduleLocation", "priority"]
